# SPDX-License-Identifier: MIT

import os

import pytest
from packaging.version import Version
from sh.contrib import git
from versioning import version

from . import gitdir, tmpdir


def _format_commit_message(message: str) -> str:
    return "\n".join(map(str.strip, message.strip().split("\n")))


def test_project_s_version_is_set():
    from versioning import __version__

    Version(__version__)


def test_version_not_in_git_repository(tmpdir):
    os.chdir(tmpdir)
    with pytest.raises(RuntimeError, match="not a git work tree"):
        version()


def test_version_default(gitdir):
    os.chdir(gitdir)
    assert version() == Version("0.0.0")


def test_version_at_tag(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    assert version() == Version("1.0.0")


def test_version_patch(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: patch
            """
        ),
    )
    assert version() == Version("1.0.1")


def test_version_minor(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: minor
            """
        ),
    )
    assert version() == Version("1.1.0")


def test_version_major(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: major
            """
        ),
    )
    assert version() == Version("2.0.0")


def test_version_major_minor_patch(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: major
            """
        ),
    )
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: minor
            """
        ),
    )
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: patch
            """
        ),
    )
    assert version() == Version("2.1.1")


def test_version_patch_minor_major(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: patch
            """
        ),
    )
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: minor
            """
        ),
    )
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: major
            """
        ),
    )
    assert version() == Version("2.0.0")


def test_version_patch_minor(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: patch
            """
        ),
    )
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: minor
            """
        ),
    )
    assert version() == Version("1.1.0")


def test_version_default_bump_is_patch(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit("--allow-empty", "--message", "lorem ipsum dolor sit amet")
    git.commit("--allow-empty", "--message", "lorem ipsum dolor sit amet")
    git.commit("--allow-empty", "--message", "lorem ipsum dolor sit amet")
    assert version() == Version("1.0.3")


def test_version_ignores_unrelated_trailers(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            Acked-by: John Smith <john.smith@example.com>
            Signoed-off-by: Jane Doe <jane.doe@example.com>
            Change-type: major
            """
        ),
    )
    assert version() == Version("1.0.1")


def test_version_typos_in_trailer_values_get_caught(gitdir):
    os.chdir(gitdir)
    git.commit("--allow-empty", "--message", "initial empty commit")
    git.tag("--annotate", "--message", "version 1.0.0", "1.0.0")
    git.commit(
        "--allow-empty",
        "--message",
        _format_commit_message(
            """
            lorem ipsum dolor sit amet

            ci-version-bump: path
            """
        ),
    )
    with pytest.raises(RuntimeError, match="patch, minor, or major"):
        version()



# SPDX-License-Identifier: MIT

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterator

import pytest
from sh.contrib import git


@pytest.fixture
def tmpdir() -> Iterator[Path]:
    with TemporaryDirectory() as directory:
        yield Path(directory)


@pytest.fixture
def gitdir(tmpdir: Path) -> Iterator[Path]:
    git.init(tmpdir)
    git.bake("-C", tmpdir).config("user.name", "Jane Doe")
    git.bake("-C", tmpdir).config("user.email", "jane.doe@example.com")
    yield tmpdir



# SPDX-License-Identifier: MIT

import os
from contextlib import contextmanager
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as importlib_metadata_version
from pathlib import Path
from typing import Iterator, Tuple

from packaging.version import Version
from sh import ErrorReturnCode
from sh.contrib import git


def _last() -> Version:
    """
    Return the last version explicitly set with a tag in the history
    """
    return Version(git.describe("--abbrev=0").stdout.decode())


def _patch(version: Version) -> Version:
    return Version(f"{version.major}.{version.minor}.{version.micro + 1}")


def _minor(version: Version) -> Version:
    return Version(f"{version.major}.{version.minor + 1}.0")


def _major(version: Version) -> Version:
    return Version(f"{version.major + 1}.0.0")


def _trailers(commit: str, *args: str) -> Iterator[Tuple[str, str]]:
    for trailer in (
        git(
            git.show("--format=format:%B", "--no-patch", commit),
            "interpret-trailers",
            "--no-divider",
            "--unfold",
            "--trim-empty",
            "--only-trailers",
            *args,
        )
        .stdout.decode()
        .strip()
        .split("\n")
    ):
        key, value = trailer.split(":", maxsplit=1)
        yield key.strip(), value.strip()


def _is_git_work_tree() -> bool:
    """
    Return True if the current directory inside a git work tree, False otherwise
    """
    try:
        return git("rev-parse", "--is-inside-work-tree").stdout.decode().strip() == "true"
    except ErrorReturnCode as exc:
        if exc.exit_code != 128:
            raise
        return False


def _version() -> Version:
    if not _is_git_work_tree():
        raise RuntimeError(f"{os.getcwd()} is not a git work tree")

    try:
        version = last = _last()
    except ErrorReturnCode as exc:
        if exc.exit_code != 128:
            raise
        return Version("0.0.0")
    for commit in git.log("--reverse", "--format=format:%H", f"{last}..").stdout.decode()[:-1].split():
        for token, value in _trailers(
            commit, "--if-missing", "add", "--if-exists", "doNothing", "--trailer", "ci-version-bump: patch"
        ):
            if token != "ci-version-bump":
                continue

            if value == "patch":
                version = _patch(version)
            elif value == "minor":
                version = _minor(version)
            elif value == "major":
                version = _major(version)
            else:
                raise RuntimeError(
                    f"Unexpected value for ci-version-bump trailer: '{value}', expected one of patch, minor, or major"
                )
    return version


@contextmanager
def _cwd(directory: Path) -> Iterator[Path]:
    cwd = os.getcwd()
    try:
        os.chdir(directory)
        yield directory
    finally:
        os.chdir(cwd)


def version(project: Path = Path(".")) -> Version:
    """
    Compute the current version of a project based on git tags and git trailers
    """
    with _cwd(project):
        return _version()


try:
    __version__ = importlib_metadata_version("simple-git-versioning")
except PackageNotFoundError:
    # package is not installed
    __version__ = str(version(Path(__file__).parent))
